/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 the PM2 team (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#ifndef __SYS_MARCEL_GLUEC_REALSYM_H__
#define __SYS_MARCEL_GLUEC_REALSYM_H__


#include <stdio.h>
#include "tbx.h"
#include "marcel_config.h"


#ifdef __MARCEL_KERNEL__
TBX_VISIBILITY_PUSH_INTERNAL


/** Symbol identifiers **/
#ifdef MARCEL_REALSYM_C
#define DEC_MA_REALSYM(retval, name, proto)	\
	retval (*MA_REALSYM(name)) proto = NULL;
#else
#define DEC_MA_REALSYM(retval, name, proto)	\
	extern retval (*MA_REALSYM(name)) proto
#endif

#define MA_REALSYM(name) ma_realsym_##name

#define DEF_MA_REALSYM(handle, name)					\
	ma_realsym_##name = dlsym(handle, #name);			\
	if (NULL == MA_REALSYM(name))	{				\
		fprintf(stderr, "realsym: %s cannot be dlsymed\n", #name); \
		return tbx_false;					\
	}

#define MA_CALL_LIBC_FUNC(rtype, func, ...)				\
	do {								\
		marcel_extlib_protect();				\
		rtype status = MA_REALSYM(func)(__VA_ARGS__);		\
		marcel_extlib_unprotect();				\
		return status;						\
	} while (0)



/** stdio **/
DEC_MA_REALSYM(int, vfprintf, (FILE *__restrict stream, const char *__restrict format, va_list ap));
DEC_MA_REALSYM(int, vprintf, (const char *__restrict format, va_list ap));
DEC_MA_REALSYM(int, vsprintf, (char *__restrict s, const char *__restrict format, va_list ap));
DEC_MA_REALSYM(int, puts, (const char *s));
DEC_MA_REALSYM(int, putchar, (int c));
DEC_MA_REALSYM(int, fputc, (int c, FILE *stream));
DEC_MA_REALSYM(int, fputs, (const char *s, FILE *stream));
DEC_MA_REALSYM(char*, fgets, (char *__restrict s, int n, FILE *__restrict stream));
DEC_MA_REALSYM(size_t, fread, (void *ptr, size_t size, size_t nmemb, FILE *stream));
DEC_MA_REALSYM(size_t, fwrite, (const void *ptr, size_t size, size_t nmemb, FILE *stream));
DEC_MA_REALSYM(FILE*, fopen, (const char *__restrict filename, const char *__restrict mode));
DEC_MA_REALSYM(int, fclose, (FILE *stream));
DEC_MA_REALSYM(int, fflush, (FILE *stream));
DEC_MA_REALSYM(int, feof, (FILE *stream));
DEC_MA_REALSYM(int, vfscanf, (FILE *__restrict stream, const char *__restrict format, va_list arg));
DEC_MA_REALSYM(int, vscanf, (const char *__restrict format, va_list arg ));
DEC_MA_REALSYM(int, vsscanf, (const char *__restrict s, const char *__restrict format, va_list arg ));

/** exec **/
DEC_MA_REALSYM(int, execve, (const char *name, char *const argv[], char *const envp[]));
DEC_MA_REALSYM(int, execv, (const char *name, char *const argv[]));
DEC_MA_REALSYM(int, execvp, (const char *name, char *const argv[]));
DEC_MA_REALSYM(int, system, (const char *command));
DEC_MA_REALSYM(pid_t, fork, (void));
#if HAVE_DECL_POSIX_SPAWN
#include <spawn.h>
DEC_MA_REALSYM(int, posix_spawn, 
	       (pid_t * __restrict pid, const char *__restrict path,
		const posix_spawn_file_actions_t * file_actions,
		const posix_spawnattr_t * __restrict attrp, char *const argv[],
		char *const envp[]));
DEC_MA_REALSYM(int, posix_spawnp, 
	       (pid_t * __restrict pid, const char *__restrict file,
		const posix_spawn_file_actions_t * file_actions,
		const posix_spawnattr_t * __restrict attrp, char *const argv[],
		char *const envp[]));
#endif

/** memory **/
DEC_MA_REALSYM(void*, calloc, (size_t nmemb, size_t size));
DEC_MA_REALSYM(void*, malloc, (size_t size));
DEC_MA_REALSYM(void*, valloc, (size_t size));
DEC_MA_REALSYM(void, free, (void *ptr));
DEC_MA_REALSYM(void*, realloc, (void *ptr, size_t size));
DEC_MA_REALSYM(int, brk, (void *addr));
DEC_MA_REALSYM(void*, sbrk, (intptr_t increment));
DEC_MA_REALSYM(void*, mmap, (void *addr, size_t length, int prot, int flags, int fd, off_t offset));
DEC_MA_REALSYM(int, munmap, (void *addr, size_t length));
DEC_MA_REALSYM(char*, strdup, (const char *s));
#if HAVE_DECL_MEMALIGN
DEC_MA_REALSYM(void*, memalign, (size_t boundary, size_t size));
#endif
#if HAVE_DECL_POSIX_MEMALIGN
DEC_MA_REALSYM(int, posix_memalign, (void **memptr, size_t alignment, size_t size));
#endif


/** Internal interface **/
tbx_bool_t ma_realsym_init(void);
void ma_realsym_exit(void);


TBX_VISIBILITY_POP
#endif /** __MARCEL_KERNEL__ **/


#endif /** __SYS_MARCEL_GLUEC_REALSYM_H__ **/
