/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 the PM2 team (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#define MARCEL_REALSYM_C
#include <dlfcn.h>
#include "sys/marcel_gluec_stdio.h"
#include "sys/marcel_gluec_realsym.h"
#include "sys/marcel_utils.h"


#if defined(HAVE_GNU_LIB_NAMES_H) && HAVE_GNU_LIB_NAMES_H
#  include <gnu/lib-names.h>
#  define LIBC_FILE LIBC_SO
#endif


static void *libc_handle = NULL;


tbx_bool_t ma_realsym_init(void)
{
	libc_handle = dlopen(LIBC_FILE, RTLD_LAZY);
	if (! libc_handle) {
		fprintf(stderr, "realsym: %s\n", dlerror());
		return tbx_false;
	}

	/** stdio **/
	DEF_MA_REALSYM(libc_handle, vfprintf);
	DEF_MA_REALSYM(libc_handle, vprintf);
	DEF_MA_REALSYM(libc_handle, vsprintf);
	DEF_MA_REALSYM(libc_handle, puts);
	DEF_MA_REALSYM(libc_handle, putchar);
	DEF_MA_REALSYM(libc_handle, fputc);
	DEF_MA_REALSYM(libc_handle, fputs);
	DEF_MA_REALSYM(libc_handle, fgets);
	DEF_MA_REALSYM(libc_handle, fread);
	DEF_MA_REALSYM(libc_handle, fwrite);
	DEF_MA_REALSYM(libc_handle, fopen);
	DEF_MA_REALSYM(libc_handle, fclose);
	DEF_MA_REALSYM(libc_handle, fflush);
	DEF_MA_REALSYM(libc_handle, feof);
	DEF_MA_REALSYM(libc_handle, vfscanf);
	DEF_MA_REALSYM(libc_handle, vscanf);
	DEF_MA_REALSYM(libc_handle, vsscanf);

	/** exec **/
	DEF_MA_REALSYM(libc_handle, execve);
	DEF_MA_REALSYM(libc_handle, execv);
	DEF_MA_REALSYM(libc_handle, execvp);
	DEF_MA_REALSYM(libc_handle, system);
	DEF_MA_REALSYM(libc_handle, fork);
#if HAVE_DECL_POSIX_SPAWN
	DEF_MA_REALSYM(libc_handle, posix_spawn);
	DEF_MA_REALSYM(libc_handle, posix_spawnp);
#endif

	/** memory **/
	DEF_MA_REALSYM(libc_handle, calloc);
	DEF_MA_REALSYM(libc_handle, malloc);
	DEF_MA_REALSYM(libc_handle, valloc);
	DEF_MA_REALSYM(libc_handle, free);
	DEF_MA_REALSYM(libc_handle, realloc);
	DEF_MA_REALSYM(libc_handle, brk);
	DEF_MA_REALSYM(libc_handle, sbrk);
	DEF_MA_REALSYM(libc_handle, mmap);
	DEF_MA_REALSYM(libc_handle, munmap);
	DEF_MA_REALSYM(libc_handle, strdup);
#if HAVE_DECL_MEMALIGN
	DEF_MA_REALSYM(libc_handle, memalign);
#endif
#if HAVE_DECL_POSIX_MEMALIGN
	DEF_MA_REALSYM(libc_handle, posix_memalign);
#endif

	return tbx_true;
}

void ma_realsym_exit(void)
{
	if (libc_handle) {
		dlclose(libc_handle);
		libc_handle = NULL;
	}
}
